@@ -3,12 +3,11 @@
 
   @author Johan Hedberg <johan.hedberg@nokia.com>
 
-  Copyright (C) 2004 Nokia. All rights reserved.
+  Copyright (C) 2006 Nokia Corporation. All rights reserved.
 
   This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License as published by
-  the Free Software Foundation; either version 2 of the License, or
-  (at your option) any later version.
+  it under the terms of the GNU General Public License, version 2, as
+  published by the Free Software Foundation.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -17,13 +16,14 @@
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 
 */
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <time.h>
 #include <errno.h>
 #include <stdint.h>
 #include <glib.h>
@@ -32,7 +32,6 @@
 
 #include "log.h"
 #include "dbus.h"
-#include "gconf.h"
 #include "bt-dbus.h"
 #include "dbus-helper.h"
 #include "bt-utils.h"
@@ -42,6 +41,8 @@
 #include "btcond-rfcomm.h"
 #include "btcond-bt.h"
 
+#define STREQ(a,b) (strcasecmp((a), (b)) == 0)
+
 #define SAP_MIN_KLEN 64
 
 #define HCITOOL "hcitool"
@@ -52,7 +53,6 @@
 /** Bluetooth device information */
 typedef struct {
     char            *bda;            /**< BT address */
-    uint16_t         clock;          /**< BT clock offset */
 
     GSList          *services;       /**< List of BtService elements */
 
@@ -62,6 +62,7 @@ typedef struct {
     gboolean         need_update;    /**< Update at next chance */
     gboolean         auto_update;    /**< If FALSE, skip in process_queue() */
 
+    GSList          *wanted_svcs;    /**< Services to be queried via SDP */
     DBusPendingCall *pending_update; /**< Waiting for SDP reply */
 
     GSList          *notify_list;    /**< List of functions to call on update */
@@ -78,7 +79,7 @@ typedef struct {
 
 typedef struct {
     BdInfo      *dev;
-    gchar       *service;
+    BtService   *svc;
     uint8_t      channel;
     gchar       *owner;      /**< D-BUS service which requested the connection */
     btcon_cb_t   func;
@@ -96,10 +97,13 @@ static GSList *devs = NULL;
 
 static gboolean sdp_in_progress;
 
-static void _schedule_capability_update(BdInfo *dev, btsdp_cb_t func,
+static void _schedule_capability_update(BdInfo *dev, btsdp_cb_t func, const char *svc,
                                         gpointer user_data, GFreeFunc free_data);
 
-static void free_pending_info(BdInfo *dev) {
+static void bt_owner_disconnect(const char *name, RfcommDev *node);
+
+static void free_pending_info(BdInfo *dev)
+{
     GSList *l;
 
     for (l = dev->notify_list; l != NULL; l = l->next) {
@@ -114,69 +118,59 @@ static void free_pending_info(BdInfo *dev) {
     dev->notify_list = NULL;
 }
 
-static BtService *bt_service_dup(BtService *svc) {
-    BtService *s;
-
-    s = g_new0(BtService, 1);
-    s->bda = svc->bda;
-    s->name = g_strdup(svc->name);
-    s->channel = svc->channel;
-
-    return s;
+static void bt_service_ref(BtService *svc)
+{
+    svc->__ref += 1;
 }
 
-static void free_service(BtService *svc, gpointer user_data) {
-    /* We don't free svc->bda since it is owned by the BdInfo struct */
-    g_free(svc->name);
-    g_free(svc);
+static void bt_service_unref(BtService *svc)
+{
+    svc->__ref -= 1;
+    if (!svc->__ref) {
+        g_free(svc->name);
+        g_free(svc->dscr);
+        g_free(svc);
+    }
 }
-
-static void free_rfcomm_dev(RfcommDev *dev) {
+                
+static void free_rfcomm_dev(RfcommDev *dev, gboolean complete)
+{
     if (dev->io) {
         rfcomm_disconnect(dev->io, dev->id, NULL);
         g_source_remove(dev->io_id);
         send_dbus_rfcomm_status(dev->svc->name, dev->bd->bda, "disconnected");
     }
-    else
+    else if (complete)
         rfcomm_release(dev->id, NULL);
-    free_service(dev->svc, NULL);
-    g_free(dev->owner);
+    bt_service_unref(dev->svc);
+    if (dev->owner) {
+        if (complete)
+            remove_name_listener(get_dbus_connection(), dev->owner,
+                    (name_cb)bt_owner_disconnect, dev);
+        g_free(dev->owner);
+    }
     g_free(dev);
 }
 
-static void free_service_list(GSList *services) {
-    g_slist_foreach(services, (GFunc)free_service, NULL);
+static void free_service_list(GSList *services)
+{
+    g_slist_foreach(services, (GFunc)bt_service_unref, NULL);
     g_slist_free(services);
 }
 
-static void free_rfcomm_dev_list(GSList *devlist) {
-    g_slist_foreach(devlist, (GFunc)free_rfcomm_dev, NULL);
+static void free_rfcomm_dev_list(GSList *devlist)
+{
+    g_slist_foreach(devlist, (GFunc)free_rfcomm_dev, (gpointer)TRUE);
     g_slist_free(devlist);
 }
 
-#if 0
-static void free_dev(BdInfo *dev) {
-    if (dev->bda)
-        g_free(dev->bda);
-    if (dev->services)
-        free_service_list(dev->services);
-    if (dev->connected)
-        free_rfcomm_dev_list(dev->connected);
-    if (dev->bound)
-        free_rfcomm_dev_list(dev->bound);
-    if (dev->pending_update)
-        dbus_pending_call_cancel(dev->pending_update);
-    if (dev->notify_list)
-        free_pending_info(dev);
-    g_free(dev);
-}
-#endif 
-
-static gint cmp_dev(BdInfo *dev, const gchar *bda) {
+static gint cmp_dev(BdInfo *dev, const gchar *bda)
+{
     return g_ascii_strcasecmp(dev->bda, bda);
 }
 
-static BdInfo *create_new_device(const char *bda) {
+static BdInfo *create_new_device(const char *bda)
+{
     BdInfo *bd;
 
     if (!bda_ok(bda)) {
@@ -195,7 +189,8 @@ static BdInfo *create_new_device(const char *bda) {
     return bd;
 }
 
-static BdInfo *get_dev(const gchar *bda) {
+static BdInfo *get_dev(const gchar *bda)
+{
     GSList *element;
 
     element = g_slist_find_custom(devs, bda, (GCompareFunc)cmp_dev);
@@ -205,22 +200,16 @@ static BdInfo *get_dev(const gchar *bda) {
         return element->data;
 }
 
-static BdInfo *get_my_dev(void) {
-    BdInfo *bd;
-    gchar *bda;
-
-    bda = get_preferred_bda();
-    if (bda == NULL)
-        return NULL;
-
-    bd = get_dev(bda);
-
-    g_free(bda);
-
-    return bd;
+static void bt_owner_disconnect(const char *name, RfcommDev *node)
+{
+    debug("D-BUS owner for (%s, %s) exited, disconnecting service.",
+            node->svc->name, node->bd->bda);
+    node->bd->connected = g_slist_remove(node->bd->connected, node);
+    free_rfcomm_dev(node, FALSE);
 }
 
-static void _expire_dev_capability(BdInfo *dev) {
+static void _expire_dev_capability(BdInfo *dev)
+{
     dev->need_update = TRUE;
 }
 
@@ -228,27 +217,32 @@ static gint cmp_service_name(BtService *svc, const char *name) {
     return strcmp(svc->name, name);
 }
 
-static gint cmp_service_chan(BtService *svc, gpointer data) {
+static gint cmp_service_chan(BtService *svc, gpointer data)
+{
     long int channel = GPOINTER_TO_INT(data);
     return svc->channel - (uint8_t)channel;
 }
 
-static gint cmp_dev_service_name(RfcommDev *dev, const char *name) {
+static gint cmp_dev_service_name(RfcommDev *dev, const char *name)
+{
     g_assert(dev->svc != NULL);
     return strcmp(dev->svc->name, name);
 }
 
-static gint cmp_dev_service_chan(RfcommDev *dev, long int *channel) {
+static gint cmp_dev_service_chan(RfcommDev *dev, long int *channel)
+{
     g_assert(dev->svc != NULL);
     return dev->svc->channel - (uint8_t)(*channel);
 }
 
-static gint cmp_dev_id(RfcommDev *dev, uint8_t *id) {
+static gint cmp_dev_id(RfcommDev *dev, uint8_t *id)
+{
     return dev->id - *id;
 }
 
 static RfcommDev *find_rfcomm_dev(GSList *devlist,
-                                  const char *name, long int ch) {
+                                  const char *name, long int ch)
+{
     GSList *l;
 
     if (ch > 255 || ch < 0)
@@ -256,14 +250,15 @@ static RfcommDev *find_rfcomm_dev(GSList *devlist,
 
     for (l = devlist; l != NULL; l = l->next) {
         RfcommDev *dev = l->data;
-        if (g_str_equal(name, dev->svc->name) && dev->svc->channel == (uint8_t)ch)
+        if (STREQ(name, dev->svc->name) && dev->svc->channel == (uint8_t)ch)
             return dev;
     }
 
     return NULL;
 }
 
-static RfcommDev *find_rfcomm_dev_id(GSList *devlist, uint8_t id) {
+static RfcommDev *find_rfcomm_dev_id(GSList *devlist, uint8_t id)
+{
     GSList *match;
 
     match = g_slist_find_custom(devlist, &id,
@@ -275,7 +270,8 @@ static RfcommDev *find_rfcomm_dev_id(GSList *devlist, uint8_t id) {
         return NULL;
 }
 
-static RfcommDev *find_rfcomm_dev_ch(GSList *devlist, long int ch) {
+static RfcommDev *find_rfcomm_dev_ch(GSList *devlist, long int ch)
+{
     GSList *match;
 
     if (ch > 255 || ch < 0)
@@ -290,10 +286,11 @@ static RfcommDev *find_rfcomm_dev_ch(GSList *devlist, long int ch) {
         return NULL;
 }
 
-static RfcommDev *find_rfcomm_dev_name(GSList *devlist, const char *name) {
+static RfcommDev *find_rfcomm_dev_name(GSList *devlist, const char *name)
+{
     GSList *match;
 
-    if (isdigit(name[0]))
+    if (isdigit(name[0]) && strlen(name) < 4)
         return find_rfcomm_dev_ch(devlist, strtol(name, NULL, 0));
     else
         match = g_slist_find_custom(devlist, name,
@@ -305,30 +302,46 @@ static RfcommDev *find_rfcomm_dev_name(GSList *devlist, const char *name) {
         return NULL;
 }
 
+static time_t calculate_expiry_time(uint32_t ttl)
+{
+    /* Give minimum 10s lifetime for services */
+    if (ttl == 0)
+        ttl = 10;
+
+    return time(NULL) + ttl;
+}
+
 /** Add service to list of services of a device
- * RFCOMM device to the service
  * @param dev     Device to operate on
+ * @param type    'r' for RFCOMM 'l' for L2CAP
  * @param name    Name of the service (eg. "DUN")
  * @param channel RFCOMM channel the service is on
  * @returns The newly created BtService
  */
-static BtService *add_service(BdInfo *dev, const char *name, uint8_t channel) {
+static BtService *add_service(BdInfo *dev, uint8_t type, const char *name,
+                              uint8_t channel, uint32_t ttl, const char *dscr)
+{
     BtService *service;
 
-    debug("%s: adding %s, channel %u", dev->bda, name, channel);
+    debug("%s: adding %s, channel %u, ttl %d, \"%s\"", dev->bda, name, channel, ttl, dscr);
 
     service = g_new0(BtService, 1);
     service->bda     = dev->bda;
+    service->type    = type;
+    service->expiry  = calculate_expiry_time(ttl);
     service->name    = g_strdup(name);
     service->channel = channel;
+    service->dscr    = g_strdup(dscr);
 
     dev->services = g_slist_append(dev->services, service);
 
+    bt_service_ref(service);
+
     return service;
 }
 
-static BtService *bt_get_service(BdInfo *dev,
-                                  const char *name, long int ch) {
+static BtService *bt_get_service(BdInfo *dev, const char *name, long int ch)
+{
     GSList *l;
 
     if (ch < 0 || ch > 255)
@@ -337,14 +350,15 @@ static BtService *bt_get_service(BdInfo *dev,
     for (l = dev->services; l != NULL; l = l->next) {
         BtService *svc = l->data;
 
-        if (g_str_equal(svc->name, name) && (uint8_t)ch == svc->channel)
+        if (STREQ(svc->name, name) && (uint8_t)ch == svc->channel)
             return svc;
     }
 
     return NULL;
 }
 
-static BtService *bt_get_service_ch(BdInfo *dev, long int ch, gboolean create) {
+static BtService *bt_get_service_ch(BdInfo *dev, long int ch, gboolean create)
+{
     GSList *match;
 
     if (ch < 0 || ch > 255)
@@ -355,30 +369,38 @@ static BtService *bt_get_service_ch(BdInfo *dev, long int ch, gboolean create) {
 
     if (match)
         return match->data;
-    else if (create) {
+
+    if (create) {
         char svc[4];
         snprintf(svc, sizeof(svc), "%d", (int)ch);
-        return add_service(dev, svc, ch);
+        return add_service(dev, 'r', svc, ch, 0, "");
     }
 
     return NULL;
 }
 
-static BtService *bt_get_service_name(BdInfo *dev, const char *name) {
-    GSList *match;
+static BtService *get_rfcomm_service_by_name(BdInfo *dev, const char *name)
+{
+    BtService *svc = NULL;
 
-    if (isdigit(name[0]))
-        return bt_get_service_ch(dev, strtol(name, NULL, 0), TRUE);
-    else
+    if (isdigit(name[0]) && strlen(name) < 4)
+        svc = bt_get_service_ch(dev, strtol(name, NULL, 0), TRUE);
+    else {
+        GSList *match;
         match = g_slist_find_custom(dev->services, name,
-                                    (GCompareFunc)cmp_service_name);
-    if (match)
-        return match->data;
-    else
-        return NULL;
+                (GCompareFunc)cmp_service_name);
+        if (match)
+            svc = match->data;
+    }
+
+    if (svc && svc->type != 'r')
+        svc = NULL;
+
+    return svc;
 }
 
-static void _reset_capability(BdInfo *dev) {
+static void _reset_capability(BdInfo *dev)
+{
     if (dev->services != NULL) {
         free_service_list(dev->services);
         dev->services = NULL;
@@ -391,47 +413,36 @@ static void _reset_capability(BdInfo *dev) {
     }
 }
 
-static gboolean disconnect_cb(GIOChannel *io, GIOCondition cond, RfcommDev *node) {
+static gboolean disconnect_cb(GIOChannel *io, GIOCondition cond, RfcommDev *node)
+{
     debug("Service %s for device %s was disconnected",
             node->svc->name, node->bd->bda);
     node->bd->connected = g_slist_remove(node->bd->connected, node);
-    free_rfcomm_dev(node);
+    free_rfcomm_dev(node, TRUE);
     return FALSE;
 }
 
-static void connect_cb(gint dev_id, GIOChannel *io, GError *err, ConnectData *data) {
+static void connect_cb(gint dev_id, GIOChannel *io, GError *err, ConnectData *data)
+{
     GError *local_err = NULL;
-    gboolean removed = FALSE;
 
     g_assert((dev_id < 0 && io == NULL) || (dev_id >= 0 && io != NULL));
 
-    debug("connect_cb: (%s, %s): %s", data->dev->bda, data->service,
+    debug("connect_cb: (%s, %s): %s", data->dev->bda, data->svc->name,
           dev_id < 0 ? "connect failed" : "connected"); 
 
-    /* Check for removed device while connecting */
-    if (g_slist_find(devs, data->dev) == NULL) {
-        error("Device or service was removed while connecting");
-        removed = TRUE;
-    }
-    /* Check for removed service while connecting */
-    if (!removed && !bt_get_service(data->dev, data->service, data->channel)) {
-        error("Service %s was removed while connecting", data->service);
-        removed = TRUE;
-    }
     /* Check if the calling D-BUS name is still present */
     if (data->owner &&
-        !dbus_bus_service_exists(get_dbus_connection(), data->owner, NULL)) {
+            !dbus_bus_name_has_owner(get_dbus_connection(), data->owner, NULL)) {
         error("Connection initiator (%s) exited", data->owner);
-        removed = TRUE;
-    }
-
-    if (removed && dev_id >= 0) {
-        rfcomm_disconnect(io, dev_id, NULL);
-        dev_id = -1;
-        g_set_error(&local_err, BT_COND_ERROR,
+        if (dev_id >= 0) {
+            rfcomm_disconnect(io, dev_id, NULL);
+            dev_id = -1;
+            g_set_error(&local_err, BT_COND_ERROR,
                     BT_COND_ERROR_FAILED,
-                    "Device or service doesn't exist anymore");
-        err = local_err;
+                    "Connection initiator (%s) exited", data->owner);
+            err = local_err;
+        }
     }
 
     if (dev_id >= 0) {
@@ -439,11 +450,15 @@ static void connect_cb(gint dev_id, GIOChannel *io, GError *err, ConnectData *da
 
         node = g_new0(RfcommDev, 1);
         node->bd  = data->dev;
-        node->svc = bt_service_dup(bt_get_service(data->dev, data->service, data->channel));
+        node->svc = data->svc;
+        bt_service_ref(node->svc);
         node->id  = dev_id;
         node->io  = io;
-        if (data->owner)
+        if (data->owner) {
             node->owner = g_strdup(data->owner);
+            add_name_listener(get_dbus_connection(), node->owner,
+                    (name_cb)bt_owner_disconnect, node);
+        }
 
         node->io_id = g_io_add_watch(io, G_IO_ERR | G_IO_HUP, (GIOFunc)disconnect_cb, node);
 
@@ -457,15 +472,16 @@ static void connect_cb(gint dev_id, GIOChannel *io, GError *err, ConnectData *da
     if (data->data && data->free_data)
         data->free_data(data->data);
 
-    if (local_err != NULL)
+    if (local_err)
         g_error_free(local_err);
 
-    g_free(data->service);
+    bt_service_unref(data->svc);
     g_free(data->owner);
     g_free(data);
 } 
 
-static void _bt_disconnect(BdInfo *dev, gboolean clear_bd_info) {
+static void _bt_disconnect(BdInfo *dev, gboolean clear_bd_info)
+{
     if (connection_status(dev->bda) == CONN_STATUS_CONNECTED) {
         gchar *cmd;
 
@@ -489,7 +505,8 @@ static void _bt_disconnect(BdInfo *dev, gboolean clear_bd_info) {
         _reset_capability(dev);
 }
 
-void bt_disconnect(const gchar *bda, gboolean clear_bd_info) {
+void bt_disconnect(const gchar *bda, gboolean clear_bd_info)
+{
     BdInfo *dev;
 
     /* If bda is NULL, disconnect all devices */
@@ -512,7 +529,8 @@ void bt_disconnect(const gchar *bda, gboolean clear_bd_info) {
     }
 }
 
-static RfcommDev *bind_service(BdInfo *dev, BtService *service, GError **err) {
+static RfcommDev *bind_service(BdInfo *dev, BtService *service, GError **err)
+{
     int dev_id;
     RfcommDev *node;
 
@@ -522,14 +540,17 @@ static RfcommDev *bind_service(BdInfo *dev, BtService *service, GError **err) {
     node = g_new0(RfcommDev, 1);
     node->bd  = dev;
     node->id  = dev_id;
-    node->svc = bt_service_dup(service);
+    node->svc = service;
+    bt_service_ref(service);
 
     dev->bound = g_slist_append(dev->bound, node);
 
     return node;
 }
 
-static void handle_btsdp_reply(DBusMessage *reply, BdInfo *dev) {
+static void handle_btsdp_reply(DBusMessage *reply, BdInfo *dev)
+{
+    const char *bda;
     DBusMessageIter iter;
 
     debug("handle_btsdp_reply()");
@@ -538,27 +559,51 @@ static void handle_btsdp_reply(DBusMessage *reply, BdInfo *dev) {
     dev->services = NULL;
 
     dbus_message_iter_init(reply, &iter);
-    while (dbus_message_iter_has_next(&iter)) {
-        uint8_t channel;
-        char *service;
 
-        if (!get_dbus_iter_args(&iter,
+    if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING)
+        return;
+
+    dbus_message_iter_get_basic(&iter, &bda);
+
+    if (dbus_message_iter_next(&iter)) {
+        DBusMessageIter list_iter;
+
+        if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_ARRAY)
+            return;
+
+        dbus_message_iter_recurse(&iter, &list_iter);
+
+        for (dbus_message_iter_recurse(&iter, &list_iter);
+                dbus_message_iter_get_arg_type(&list_iter) != DBUS_TYPE_INVALID;
+                dbus_message_iter_next(&list_iter)) {
+            uint32_t ttl;
+            uint8_t channel, type;
+            const char *service, *name;
+            BtService *svc;
+            DBusMessageIter svc_iter;
+
+            dbus_message_iter_recurse(&list_iter, &svc_iter);
+
+            if (!get_basic_args(&svc_iter,
+                                DBUS_TYPE_BYTE, &type,
                                 DBUS_TYPE_STRING, &service,
                                 DBUS_TYPE_BYTE, &channel,
-                                DBUS_TYPE_INVALID)) {
-            error("Invalid args in btsdp reply");
-            break;
+                                DBUS_TYPE_UINT32, &ttl,
+                                DBUS_TYPE_STRING, &name,
+                                DBUS_TYPE_INVALID))
+                continue;
+
+            svc = bt_get_service(dev, service, channel);
+            if (svc)
+                svc->expiry = calculate_expiry_time(ttl);
+            else
+                add_service(dev, type, service, channel, ttl, name);
         }
-
-        /* Check for existing service */
-        if (!bt_get_service(dev, service, channel))
-            add_service(dev, service, channel);
-
-        dbus_free(service);
     }
 }
 
-static void process_update_queue(void) {
+static void process_update_queue(void)
+{
     GSList *l;
 
     debug("process_update_queue()");
@@ -567,13 +612,14 @@ static void process_update_queue(void) {
         BdInfo *dev;
         dev = l->data;
         if (dev->auto_update && dev->need_update) {
-            _schedule_capability_update(dev, NULL, NULL, NULL);
+            _schedule_capability_update(dev, NULL, NULL, NULL, NULL);
             return;
         }
     }
 }
 
-static void btsdp_notify(DBusPendingCall *pending, BdInfo *dev) {
+static void btsdp_notify(DBusPendingCall *pending, BdInfo *dev)
+{
     DBusMessage *reply;
     GError *err = NULL;
     DBusError derr;
@@ -586,7 +632,7 @@ static void btsdp_notify(DBusPendingCall *pending, BdInfo *dev) {
 
     dbus_error_init(&derr);
 
-    reply = dbus_pending_call_get_reply(pending);
+    reply = dbus_pending_call_steal_reply(pending);
     if (reply == NULL) {
         debug("btsdp_reply_notify: reply timed out for %s", dev->bda);
         g_set_error(&err, BT_COND_ERROR,
@@ -602,11 +648,13 @@ static void btsdp_notify(DBusPendingCall *pending, BdInfo *dev) {
                 "%s: %s", derr.name, derr.message);
         dbus_error_free(&derr);
         dev->auto_update = FALSE;
+        dbus_message_unref(reply);
     }
     else {
         handle_btsdp_reply(reply, dev);
         dev->need_update = FALSE;
         dev->auto_update = TRUE;
+        dbus_message_unref(reply);
     }
 
     if (dev->notify_list) {
@@ -632,39 +680,48 @@ static void btsdp_notify(DBusPendingCall *pending, BdInfo *dev) {
     process_update_queue();
 }
 
-static DBusMessage *new_devcap_request(BdInfo *dev) {
+static DBusMessage *new_sdp_request(BdInfo *dev)
+{
     DBusMessage *request;
+    GSList *l;
 
     request = new_dbus_method_call(BTSDP_SERVICE,
                                    BTSDP_REQ_PATH,
                                    BTSDP_REQ_INTERFACE,
-                                   BTSDP_GET_RFCOMM_SERVICES_REQ);
+                                   BTSDP_GET_SERVICES_REQ);
+
+    append_dbus_args(request,
+                     DBUS_TYPE_STRING, &dev->bda,
+                     DBUS_TYPE_INVALID);
 
     /* Some BT stacks (e.g. the Widcomm windows stack) don't support the public
      * browse group (for some reason), so we ask for each UUID separately */
-    append_dbus_args(request,
-                     DBUS_TYPE_STRING, dev->bda,
-                     DBUS_TYPE_UINT32, (uint32_t)(dev->clock),
-                     DBUS_TYPE_STRING, "DUN",
-                     DBUS_TYPE_STRING, "FTP",
-                     DBUS_TYPE_STRING, "NFTP",
-                     DBUS_TYPE_STRING, "SAP",
-                     DBUS_TYPE_STRING, "OPP",
-                     DBUS_TYPE_STRING, "SPP",
-                     DBUS_TYPE_INVALID);
-
-    /* Make sure btsdp is auto-activated */
-    dbus_message_set_auto_activation(request, TRUE);
+    for (l = dev->wanted_svcs; l != NULL; l = l->next) {
+        const char *svc = l->data;
+        append_dbus_args(request,
+                         DBUS_TYPE_STRING, &svc,
+                         DBUS_TYPE_INVALID);
+    }
 
     return request;
 }
 
-static void _schedule_capability_update(BdInfo *dev, btsdp_cb_t func,
-                                        gpointer user_data, GFreeFunc free_data) {
+static void _schedule_capability_update(BdInfo *dev, btsdp_cb_t func, const char *svc,
+                                        gpointer user_data, GFreeFunc free_data)
+{
     DBusConnection *conn = get_dbus_connection();
     DBusPendingCall *pending;
     DBusMessage *request;
 
+    if (svc && !g_slist_find_custom(dev->wanted_svcs, svc, (GCompareFunc)strcmp)) {
+        dev->wanted_svcs = g_slist_append(dev->wanted_svcs, g_strdup(svc));
+
+        /* Search for both NFTP and FTP when NFTP is given */
+        if (g_str_equal(svc, "NFTP") && !g_slist_find_custom(dev->wanted_svcs, "FTP", (GCompareFunc)strcmp))
+                dev->wanted_svcs = g_slist_append(dev->wanted_svcs, g_strdup("FTP"));
+    }
+                
+
     if (func) {
         PendingSdpInfo *info;
         info = g_new0(PendingSdpInfo, 1);
@@ -681,12 +738,16 @@ static void _schedule_capability_update(BdInfo *dev, btsdp_cb_t func,
 
     debug("Sending SDP query to %s", dev->bda);
 
-    request = new_devcap_request(dev);
+    request = new_sdp_request(dev);
     if (!dbus_connection_send_with_reply(conn, request, &pending, BTSDP_TIMEOUT))
         die("Out of memory during dbus_connection_send_with_reply()");
     dbus_connection_flush(conn);
     dbus_message_unref(request);
 
+    g_slist_foreach(dev->wanted_svcs, (GFunc)g_free, NULL);
+    g_slist_free(dev->wanted_svcs);
+    dev->wanted_svcs = NULL;
+
     dev->pending_update = pending;
 
     sdp_in_progress = TRUE;
@@ -697,12 +758,13 @@ static void _schedule_capability_update(BdInfo *dev, btsdp_cb_t func,
         die("Out of memory during dbus_pending_call_set_notify()");
 }
 
-gboolean schedule_capability_update(const gchar *bda, btsdp_cb_t func,
+gboolean schedule_capability_update(const gchar *bda, btsdp_cb_t func, const char *svc,
                                     gpointer user_data, GFreeFunc free_data,
-                                    GError **err) {
+                                    GError **err)
+{
     BdInfo *dev;
 
-    debug("schedule_capability_update()");
+    debug("schedule_capability_update(%s, %s)", bda ? bda : "[def]", svc ? svc : "[all]");
 
     if (bt_disabled()) {
         g_set_error(err, BT_COND_ERROR,
@@ -711,10 +773,7 @@ gboolean schedule_capability_update(const gchar *bda, btsdp_cb_t func,
         return FALSE;
     }
 
-    if (bda == NULL)
-        dev = get_my_dev();
-    else
-        dev = get_dev(bda);
+    dev = get_dev(bda);
 
     if (dev == NULL) {
         g_set_error(err, BT_COND_ERROR,
@@ -726,7 +785,7 @@ gboolean schedule_capability_update(const gchar *bda, btsdp_cb_t func,
     /* Allow automatic updates */
     dev->auto_update = TRUE;
 
-    _schedule_capability_update(dev, func, user_data, free_data);
+    _schedule_capability_update(dev, func, svc, user_data, free_data);
 
     return TRUE;
 }
@@ -734,7 +793,8 @@ gboolean schedule_capability_update(const gchar *bda, btsdp_cb_t func,
 gboolean connect_service(const char *bda, const char *name, const char *owner,
                          gboolean auth, gboolean encr, const char *role,
                          btcon_cb_t func, gpointer user_data, GFreeFunc free_data,
-                         GError **err) {
+                         GError **err)
+{
     ConnectData *data;
     BtService *svc;
     BdInfo *bd;
@@ -753,10 +813,7 @@ gboolean connect_service(const char *bda, const char *name, const char *owner,
         return FALSE;
     }
 
-    if (bda == NULL)
-        bd = get_my_dev();
-    else
-        bd = get_dev(bda);
+    bd = get_dev(bda);
 
     if (bd == NULL) {
         g_set_error(err, BT_COND_ERROR,
@@ -765,7 +822,7 @@ gboolean connect_service(const char *bda, const char *name, const char *owner,
         return FALSE;
     }
 
-    svc = bt_get_service_name(bd, name);
+    svc = get_rfcomm_service_by_name(bd, name);
     if (svc == NULL) {
         g_set_error(err, BT_COND_ERROR,
                     BT_COND_ERROR_INVALID_SVC,
@@ -784,7 +841,8 @@ gboolean connect_service(const char *bda, const char *name, const char *owner,
     }
 
     data = g_new0(ConnectData, 1);
-    data->service   = g_strdup(svc->name);
+    data->svc       = svc;
+    bt_service_ref(svc);
     data->channel   = svc->channel;
     data->dev       = bd;
     if (owner)
@@ -799,9 +857,9 @@ gboolean connect_service(const char *bda, const char *name, const char *owner,
     params.encrypt = encr;
 
     if (role) {
-        if (g_str_equal(role, "slave"))
+        if (STREQ(role, "slave"))
             params.role    = ROLE_SLAVE;
-        else if (g_str_equal(role, "master"))
+        else if (STREQ(role, "master"))
             params.role    = ROLE_MASTER;
         else
             params.role    = ROLE_ANY;
@@ -809,12 +867,12 @@ gboolean connect_service(const char *bda, const char *name, const char *owner,
     else
         params.role = ROLE_ANY;
     
-    if (g_str_equal(name, "SAP"))
+    if (STREQ(name, "SAP"))
         params.min_klen = SAP_MIN_KLEN;
 
     if (!rfcomm_connect(svc->bda, svc->channel, &params,
                         (rfcomm_connect_cb)connect_cb, data, err)) {
-        g_free(data->service);
+        bt_service_unref(data->svc);
         g_free(data->owner);
         g_free(data);
         return FALSE;
@@ -823,16 +881,14 @@ gboolean connect_service(const char *bda, const char *name, const char *owner,
     return TRUE;
 }
 
-gboolean disconnect_service(const char *bda, const char *name, GError **err) {
+gboolean disconnect_service(const char *bda, const char *name, GError **err)
+{
     RfcommDev *dev;
     BdInfo *bd;
 
     debug("disconnect_service(\"%s\", \"%s\")", bda ? bda : "", name);
 
-    if (bda == NULL)
-        bd = get_my_dev();
-    else
-        bd = get_dev(bda);
+    bd = get_dev(bda);
 
     if (bd == NULL) {
         g_set_error(err, BT_COND_ERROR,
@@ -850,22 +906,20 @@ gboolean disconnect_service(const char *bda, const char *name, GError **err) {
     }
 
     bd->connected = g_slist_remove(bd->connected, dev);
-    free_rfcomm_dev(dev);
+    free_rfcomm_dev(dev, TRUE);
 
     return TRUE;
 }
 
-gboolean release_dev(const char *bda, const char *name, GError **err) {
+gboolean release_dev(const char *bda, const char *name, GError **err)
+{
     RfcommDev *dev;
     BtService *svc;
     BdInfo *bd;
 
     debug("release_dev(\"%s\", \"%s\")", bda ? bda : "", name);
 
-    if (bda == NULL)
-        bd = get_my_dev();
-    else
-        bd = get_dev(bda);
+    bd = get_dev(bda);
 
     if (bd == NULL) {
         g_set_error(err, BT_COND_ERROR,
@@ -874,7 +928,7 @@ gboolean release_dev(const char *bda, const char *name, GError **err) {
         return FALSE;
     }
 
-    svc = bt_get_service_name(bd, name);
+    svc = get_rfcomm_service_by_name(bd, name);
     if (svc == NULL) {
         g_set_error(err, BT_COND_ERROR,
                     BT_COND_ERROR_INVALID_SVC,
@@ -890,20 +944,24 @@ gboolean release_dev(const char *bda, const char *name, GError **err) {
         return FALSE;
     }
 
-    return rfcomm_release(dev->id, err);
+    if (!rfcomm_release(dev->id, err))
+        return FALSE;
+
+    bd->bound = g_slist_remove(bd->bound, dev);
+    free_rfcomm_dev(dev, FALSE);
+
+    return TRUE;
 }
 
-int bind_dev(const char *bda, const char *name, GError **err) {
+int bind_dev(const char *bda, const char *name, GError **err)
+{
     RfcommDev *dev;
     BtService *svc;
     BdInfo *bd;
 
     debug("bind_dev(\"%s\", \"%s\")", bda ? bda : "", name);
 
-    if (bda == NULL)
-        bd = get_my_dev();
-    else
-        bd = get_dev(bda);
+    bd = get_dev(bda);
 
     if (bd == NULL) {
         g_set_error(err, BT_COND_ERROR,
@@ -912,7 +970,7 @@ int bind_dev(const char *bda, const char *name, GError **err) {
         return -1;
     }
 
-    svc = bt_get_service_name(bd, name);
+    svc = get_rfcomm_service_by_name(bd, name);
     if (svc == NULL) {
         g_set_error(err, BT_COND_ERROR,
                     BT_COND_ERROR_INVALID_SVC,
@@ -936,15 +994,13 @@ int bind_dev(const char *bda, const char *name, GError **err) {
 }
 
 
-GSList *get_service_name_list(const gchar *bda, GError **err) {
+GSList *get_service_name_list(const gchar *bda, GError **err)
+{
     GSList *svc_list, *l;
     BtService *svc;
     BdInfo *dev;
 
-    if (bda == NULL)
-        dev = get_my_dev();
-    else
-        dev = get_dev(bda);
+    dev = get_dev(bda);
 
     if (dev == NULL) {
         g_set_error(err, BT_COND_ERROR,
@@ -962,20 +1018,19 @@ GSList *get_service_name_list(const gchar *bda, GError **err) {
     return svc_list;
 }
 
-void bt_hw_error(uint8_t code) {
+void bt_hw_error(uint8_t code)
+{
     error("Bluetooth HW error: 0x%02x", code);
     bt_disconnect(NULL, TRUE);
     hci_bt_down();
     hci_bt_up();
 }
 
-void expire_device_capability(const gchar *bda) {
+void expire_device_capability(const gchar *bda)
+{
     BdInfo *dev;
 
-    if (bda == NULL)
-        dev = get_my_dev();
-    else
-        dev = get_dev(bda);
+    dev = get_dev(bda);
 
     if (dev == NULL)
         return;
@@ -983,12 +1038,32 @@ void expire_device_capability(const gchar *bda) {
     _expire_dev_capability(dev);
 }
 
-gboolean need_capability_update(const gchar *bda, const gchar *svc) {
-    /* Always return true because we don't store SDP availability information */
+gboolean need_capability_update(const gchar *bda, const gchar *name)
+{
+    BtService *svc;
+    BdInfo *dev;
+
+    dev = get_dev(bda);
+
+    if (!dev)
+        return TRUE;
+
+    svc = get_rfcomm_service_by_name(dev, name);
+    if (!svc)
+        return TRUE;
+
+    if (time(NULL) < svc->expiry)
+        return FALSE;
+
+    /* Remove expired service */
+    dev->services = g_slist_remove(dev->services, svc);
+    bt_service_unref(svc);
+
     return TRUE;
 }
 
-GSList *get_bda_list(void) {
+GSList *get_bda_list(void)
+{
     GSList *l, *bda_list = NULL;
     BdInfo *dev;
 
@@ -1000,13 +1075,11 @@ GSList *get_bda_list(void) {
     return bda_list;
 }
 
-const gchar *bt_service_status(const char *bda, const char *name, GError **err) {
+const gchar *bt_service_status(const char *bda, const char *name, GError **err)
+{
     BdInfo *dev;
 
-    if (bda == NULL)
-        dev = get_my_dev();
-    else
-        dev = get_dev(bda);
+    dev = get_dev(bda);
 
     if (dev == NULL) {
         g_set_error(err, BT_COND_ERROR,
@@ -1021,7 +1094,8 @@ const gchar *bt_service_status(const char *bda, const char *name, GError **err)
         return "disconnected";
 }
 
-BtService *get_bound_service(uint8_t dev_id) {
+BtService *get_bound_service(uint8_t dev_id)
+{
     GSList *l;
 
     for (l = devs; l != NULL; l = l->next) {
@@ -1039,7 +1113,8 @@ BtService *get_bound_service(uint8_t dev_id) {
     return NULL;
 }
 
-BtService *get_service(const char *bda, const char *name, GError **err) {
+BtService *get_service(const char *bda, const char *name, GError **err)
+{
     BdInfo *dev;
     BtService *svc;
 
@@ -1052,7 +1127,7 @@ BtService *get_service(const char *bda, const char *name, GError **err) {
         return NULL;
     }
 
-    svc = bt_get_service_name(dev, name);
+    svc = get_rfcomm_service_by_name(dev, name);
     if (svc == NULL)
         g_set_error(err, BT_COND_ERROR,
                     BT_COND_ERROR_INVALID_SVC,
@@ -1061,7 +1136,8 @@ BtService *get_service(const char *bda, const char *name, GError **err) {
     return svc;
 }
 
-BtService *get_connected_service(uint8_t dev_id) {
+BtService *get_connected_service(uint8_t dev_id)
+{
     GSList *l;
 
     for (l = devs; l != NULL; l = l->next) {
@@ -1079,39 +1155,9 @@ BtService *get_connected_service(uint8_t dev_id) {
     return NULL;
 }
 
-char *bt_devname_from_id(gint dev_num, char *dev, int max_len) {
+char *bt_devname_from_id(gint dev_num, char *dev, int max_len)
+{
     snprintf(dev, max_len, "/dev/rfcomm%d", dev_num);
     return dev;
 }
 
-void bt_check_owner_disconnect(const char *owner) {
-    GSList *l;
-
-    for (l = devs; l != NULL; l = l->next) {
-        GSList *c;
-        BdInfo *bd;
-
-        g_assert(l->data != NULL);
-        bd = l->data;
-
-        for (c = bd->connected; c != NULL;) {
-            RfcommDev *dev;
-
-            g_assert(c->data != NULL);
-            dev = c->data;
-
-            c = c->next;
-
-            if (dev->owner == NULL)
-                continue;
-
-            if (strcmp(dev->owner, owner) == 0) {
-                debug("D-BUS owner for (%s, %s) exited, disconnecting service.",
-                        dev->svc->name, dev->bd->bda);
-                dev->bd->connected = g_slist_remove(dev->bd->connected, dev);
-                free_rfcomm_dev(dev);
-            }
-        }
-    }
-}
-
