@@ -3,12 +3,11 @@
 
   @author Johan Hedberg <johan.hedberg@nokia.com>
 
-  Copyright (C) 2004 Nokia. All rights reserved.
+  Copyright (C) 2006 Nokia Corporation. All rights reserved.
 
   This program is free software; you can redistribute it and/or modify
-  it under the terms of the GNU General Public License as published by
-  the Free Software Foundation; either version 2 of the License, or
-  (at your option) any later version.
+  it under the terms of the GNU General Public License, version 2, as
+  published by the Free Software Foundation.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -17,7 +16,7 @@
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 
 */
 #include <unistd.h>
@@ -38,8 +37,6 @@
 
 #include <glib.h>
 
-#include <bttools-dbus.h>
-
 #ifdef HAVE_CONFIG_H
 # include "config.h"
 #endif
@@ -67,7 +64,6 @@ typedef struct {
     int                 dd;
     uint8_t             min_klen;
     uint8_t             role;
-    uint16_t            handle;
     gboolean            auth;
     gboolean            encrypt;
     gboolean            canceled;
@@ -106,7 +102,8 @@ static void rfcomm_create_dev(rfcomm_cb_info *info);
 static void rfcomm_connect_finalize(int fd, rfcomm_cb_info *info, GError *err);
 
 
-static rfcomm_cb_info *find_pending_connect(const char *bda, uint8_t ch) {
+static rfcomm_cb_info *find_pending_connect(const char *bda, uint8_t ch)
+{
     bdaddr_t ba;
     GSList *l;
 
@@ -123,7 +120,8 @@ static rfcomm_cb_info *find_pending_connect(const char *bda, uint8_t ch) {
 }
 
 
-static gboolean set_socket_params(int sock, uint32_t opts, GError **err) {
+static gboolean set_socket_params(int sock, uint32_t opts, GError **err)
+{
     long arg;
 
     if (opts) {
@@ -173,7 +171,8 @@ static gboolean set_socket_params(int sock, uint32_t opts, GError **err) {
 
 #ifdef USE_NOKIA_HCI
 static gboolean get_key_length(int dd, uint16_t handle,
-                               uint8_t *klen, GError **err) {
+                               uint8_t *klen, GError **err)
+{
     evt_enc_key_len_read rp;
     read_enc_key_len cp;
     struct hci_request rq;
@@ -213,81 +212,10 @@ static gboolean get_key_length(int dd, uint16_t handle,
 }
 #endif
 
-static void btpair_cb(DBusPendingCall *pending, rfcomm_cb_info *info) {
-    GError *err = NULL;
-    DBusMessage *reply;
-    DBusError derr;
-
-    debug("btpair_cb");
-
-    if (info->canceled) {
-        g_set_error(&err, BT_COND_ERROR,
-                    BT_COND_ERROR_CANCELLED,
-                    "Connection was canceled");
-        rfcomm_connect_finalize(-1, info, err);
-        dbus_pending_call_unref(pending);
-        return;
-    }
-
-    reply = dbus_pending_call_get_reply(pending);
-
-    dbus_error_init(&derr);
-
-    if (reply == NULL) {
-        g_set_error(&err, BT_COND_ERROR,
-                    BT_COND_ERROR_CONNECT_FAILED,
-                    "No reply from btpair");
-        rfcomm_connect_finalize(-1, info, err);
-        dbus_pending_call_unref(pending);
-        return;
-    }
-    else if (dbus_set_error_from_message(&derr, reply)) {
-        gint e;
-
-        debug("btpair failed: %s (%s)", derr.message, derr.name);
-
-        if (dbus_error_has_name(&derr, BTPAIR_ERROR_AUTH_FAILED))
-            e = BT_COND_ERROR_AUTH_FAILED;
-        else
-            e = BT_COND_ERROR_CONNECT_FAILED;
-
-        g_set_error(&err, BT_COND_ERROR, e,
-                    "%s: %s", derr.name, derr.message);
-        dbus_error_free(&derr);
-        rfcomm_connect_finalize(-1, info, err);
-        dbus_pending_call_unref(pending);
-        return;
-    }
-
-    if (info->min_klen) {
-#ifdef USE_NOKIA_HCI
-        uint8_t klen;
-        if (!get_key_length(info->dd, htobs(info->handle), &klen, NULL))
-            error("Could not get encryption key length");
-        else {
-            debug("Got encryption key length %d for connection handle %d", klen,
-                    htobs(info->handle));
-            if (klen < info->min_klen) {
-                g_set_error(&err, BT_COND_ERROR,
-                            BT_COND_ERROR_SMALL_CRYPT_KEY,
-                            "Encryption key is less than required minimum: %d < %d",
-                            klen, info->min_klen);
-                rfcomm_connect_finalize(-1, info, err);
-                dbus_pending_call_unref(pending);
-                return;
-            }
-        }
-#else
-        (void)(0);
-#endif
-    }
-
-    rfcomm_create_dev(info);
-    dbus_pending_call_unref(pending);
-}
-
-static gboolean setup_conn(rfcomm_cb_info *info, GError **err) {
-    struct hci_conn_info_req *cr;
+static gboolean setup_conn(rfcomm_cb_info *info, GError **err)
+{
+    struct hci_conn_info_req *cr = NULL;
+    gboolean ret = FALSE;
 
     info->dd = hci_open_dev(BT_DEV_ID);
     if (info->dd < 0) {
@@ -295,7 +223,7 @@ static gboolean setup_conn(rfcomm_cb_info *info, GError **err) {
         g_set_error(err, BT_COND_ERROR,
                     BT_COND_ERROR_CONNECT_FAILED,
                     "hci_open_dev(%d): %s", BT_DEV_ID, g_strerror(errno));
-        return FALSE;
+        goto out;
     }
 
     cr = g_malloc0((sizeof(*cr) + sizeof(struct hci_conn_info)));
@@ -306,13 +234,9 @@ static gboolean setup_conn(rfcomm_cb_info *info, GError **err) {
         g_set_error(err, BT_COND_ERROR,
                     BT_COND_ERROR_CONNECT_FAILED,
                     "ioctl(HCIGETCONNINFO): %s", g_strerror(errno));
-        g_free(cr);
-        return FALSE;
+        goto out;
     }
 
-    info->handle = cr->conn_info->handle;
-    g_free(cr);
-
     if (info->role != ROLE_ANY) {
         if (hci_switch_role(info->dd, &info->raddr.rc_bdaddr, info->role, 10000) < 0) {
             error("hci_switch_role(%s): %s",
@@ -322,45 +246,38 @@ static gboolean setup_conn(rfcomm_cb_info *info, GError **err) {
         }
     }
 
-    if (info->auth) {
-        DBusConnection *conn;
-        DBusPendingCall *pending;
-        DBusMessage *msg;
-
-        msg = new_dbus_method_call(BTPAIR_SERVICE,
-                                   BTPAIR_PATH,
-                                   BTPAIR_INTERFACE,
-                                   BTPAIR_REQ_PAIR_HANDLE);
-        append_dbus_args(msg,
-                         DBUS_TYPE_UINT32, (uint32_t)(info->handle),
-                         DBUS_TYPE_BOOLEAN, info->encrypt,
-                         DBUS_TYPE_INVALID);
-        dbus_message_set_auto_activation(msg, TRUE);
-
-        conn = get_dbus_connection();
-
-        if (!dbus_connection_send_with_reply(conn, msg, &pending,
-                                             PAIR_TIMEOUT))
-            error("Sending method call to btpair failed");
-
-        dbus_connection_flush(conn);
-        dbus_message_unref(msg);
-
-        if (!dbus_pending_call_set_notify(pending,
-                                     (DBusPendingCallNotifyFunction)btpair_cb,
-                                     info, NULL))
-            die("Out of memory during dbus_pending_call_set_notify()");
-
-        return TRUE;
+#ifdef USE_NOKIA_HCI
+    if (info->min_klen) {
+        uint8_t klen;
+        if (!get_key_length(info->dd, htobs(cr->conn_info->handle), &klen, NULL))
+            error("Could not get encryption key length");
+        else {
+            debug("Got encryption key length %d for connection handle %d", klen,
+                    htobs(cr->conn_info->handle));
+            if (klen < info->min_klen) {
+                g_set_error(&err, BT_COND_ERROR,
+                            BT_COND_ERROR_SMALL_CRYPT_KEY,
+                            "Encryption key is less than required minimum: %d < %d",
+                            klen, info->min_klen);
+                goto out;
+            }
+        }
     }
+#endif
 
     rfcomm_create_dev(info);
 
-    return TRUE;
+    ret = TRUE;
+
+out:
+    g_free(cr);
+
+    return ret;
 }
 
 /* Connection setup stuff which doesn't block can go into this function */
-static void rfcomm_connect_finalize(int fd, rfcomm_cb_info *info, GError *err) {
+static void rfcomm_connect_finalize(int fd, rfcomm_cb_info *info, GError *err)
+{
     pending_connects = g_slist_remove(pending_connects, info);
 
     if (fd < 0) {
@@ -395,7 +312,8 @@ static void rfcomm_connect_finalize(int fd, rfcomm_cb_info *info, GError *err) {
         g_error_free(err);
 }
 
-static gboolean rfcomm_open_cb(rfcomm_cb_info *info) {
+static gboolean rfcomm_open_cb(rfcomm_cb_info *info)
+{
     GError *err = NULL;
     int fd;
 
@@ -425,7 +343,8 @@ static gboolean rfcomm_open_cb(rfcomm_cb_info *info) {
     return FALSE;
 }
 
-static void rfcomm_create_dev(rfcomm_cb_info *info) {
+static void rfcomm_create_dev(rfcomm_cb_info *info)
+{
     struct rfcomm_dev_req req;
     GError *err = NULL;
     int sk, fd;
@@ -461,9 +380,11 @@ static void rfcomm_create_dev(rfcomm_cb_info *info) {
         rfcomm_connect_finalize(fd, info, NULL);
 }
 
-static gboolean sock_io_cb(GIOChannel *chan, GIOCondition cond, rfcomm_cb_info *info) {
+static gboolean sock_io_cb(GIOChannel *chan, GIOCondition cond, rfcomm_cb_info *info)
+{
     GError *err = NULL;
-    int alen, sk, ret;
+    int sk, ret;
+    size_t alen;
     socklen_t len;
 
     if (cond != G_IO_OUT) {
@@ -520,7 +441,8 @@ static gboolean sock_io_cb(GIOChannel *chan, GIOCondition cond, rfcomm_cb_info *
     return FALSE;
 }
 
-gboolean rfcomm_release(int dev_num, GError **err) {
+gboolean rfcomm_release(int dev_num, GError **err)
+{
     struct rfcomm_dev_req req;
 
     g_assert(rfcomm_ctl >= 0);
@@ -540,7 +462,8 @@ gboolean rfcomm_release(int dev_num, GError **err) {
     return TRUE;
 }
 
-gboolean rfcomm_bind(const char *bda, uint8_t ch, int *node, GError **err) {
+gboolean rfcomm_bind(const char *bda, uint8_t ch, int *node, GError **err)
+{
     struct rfcomm_dev_req req;
 
     g_assert(rfcomm_ctl >= 0);
@@ -566,7 +489,8 @@ gboolean rfcomm_bind(const char *bda, uint8_t ch, int *node, GError **err) {
 }
 
 gboolean rfcomm_connect(const char *bda, uint8_t ch, rfcomm_params *params,
-                        rfcomm_connect_cb cb, gpointer user_data, GError **err) {
+                        rfcomm_connect_cb cb, gpointer user_data, GError **err)
+{
     struct sockaddr_rc laddr, raddr;
     rfcomm_cb_info *info;
     int sk;
@@ -664,7 +588,8 @@ gboolean rfcomm_connect(const char *bda, uint8_t ch, rfcomm_params *params,
     return TRUE;
 }
 
-gboolean rfcomm_cancel_connect(const char *bda, uint8_t ch, GError **err) {
+gboolean rfcomm_cancel_connect(const char *bda, uint8_t ch, GError **err)
+{
     rfcomm_cb_info *info;
 
     info = find_pending_connect(bda, ch);
@@ -680,13 +605,15 @@ gboolean rfcomm_cancel_connect(const char *bda, uint8_t ch, GError **err) {
     return TRUE;
 }
 
-gboolean rfcomm_disconnect(GIOChannel *sock, int dev_num, GError **err) {
+gboolean rfcomm_disconnect(GIOChannel *sock, int dev_num, GError **err)
+{
     g_io_channel_shutdown(sock, FALSE, NULL);
     g_io_channel_unref(sock);
     return rfcomm_release(dev_num, err);
 }
 
-gboolean rfcomm_init(GError **err) {
+gboolean rfcomm_init(GError **err)
+{
     if (rfcomm_ctl < 0) {
         rfcomm_ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM);
         if (rfcomm_ctl < 0) {
